#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>

#include "adwin_tcp.h"
#include "ADEthernet.h"
#include "UDPADwin.h"
#include "adwin-macros.h"

#ifdef _DEBUG
#include <stdio.h>
void XTrace(char *format, ...)
{
    va_list args;
    va_start(args, format);
    vprintf_s(format, args);
    va_end(args);
}
#define XTRACE XTrace
#else
#define XTRACE(...) {}
#endif

extern ADwinDevice *GetCurrentDevice ();
#define D (GetCurrentDevice ())

// ******************MG**************************

void adwin_tcp_construct(const char *ip, int16_t port, const char *passwd, int32_t timeout)
{
    struct hostent      *hp;
    unsigned long       inaddr;
    //LONGLONG g_Frequency;

    // this memset must remain at the top of the function, because the destructor checks
    // if ati.resp and ati.req contain pointers that need to be freed. If the constructor leaves
    // garbage in ati, the destructor will call free() on invalid pointers.
    memset(&D->ati, 0, sizeof(adwin_tcp_info_t));

    // printf("adwin_tcp_constructed: %s %d %s %d\n", ip, port, passwd, timeout);

    D->ati.server_addr.sin_family = AF_INET;
    D->ati.server_addr.sin_port = htons(port);
    if ((inaddr = inet_addr(ip)) != INADDR_NONE) {      // dotted decimal
        D->ati.server_addr.sin_addr.s_addr   = inaddr;
    } else {
        if ((hp = gethostbyname(ip)) == NULL) {           // by name
            PRINTDEBUG(DBG_GENERIC_DRV, ("error gethostbyname!\n"));
            return ;
        }
        D->ati.server_addr.sin_addr.s_addr   = inet_addr(hp->h_addr);
    }
    D->ati.error_number = 0;
    D->ati.firmware_version = D->ati.firmware_version_beta = 0;
    D->ati.processortype = -1;

    D->ati.initialized = 1;

    D->ati.timeout_size = 10 * 1024 * 1024;
    D->ati.req.max = SMALL_REQ_SIZE;
    D->ati.req.data.as_void = malloc(D->ati.req.max + 4); // todo: alloca?

    D->ati.req.data.as_int32[SMALL_REQ_SIZE / 4] = 0x12348765;
    D->ati.resp.max = SMALL_REQ_SIZE;
    D->ati.resp.data.as_void = malloc(D->ati.resp.max + 4); // todo alloca?
    D->ati.resp.data.as_int32[SMALL_REQ_SIZE / 4] = 0x22334455;
    D->ati.dll_version = 1;
    D->ati.timeout_ms = timeout;

    //  int32_t snd_buf = 0;

    D->ati.socket = socket(AF_INET, SOCK_STREAM, 0);
    if (D->ati.socket == -1)
        PRINTDEBUG(DBG_GENERIC_DRV, ("socket failed with error: %m\n"));



    /* disable send buffer to allow for our own timeout handling. If send buffering is not disabled,
       WSAsend returns immediately even for large writes, causing our following status read to time out! */
    int32_t bufsize = 1024 * 128;
    if (setsockopt(D->ati.socket, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof(bufsize)))
        PRINTDEBUG(DBG_GENERIC_DRV, ("setsockopt for SO_SNDBUF failed with error: %m\n"));
    if (setsockopt(D->ati.socket, SOL_SOCKET, SO_RCVBUF, &bufsize, sizeof(bufsize)))
        PRINTDEBUG(DBG_GENERIC_DRV, ("setsockopt for SO_RCVBUF failed with error: %m\n"));

    struct timeval tv;
    tv.tv_sec = D->ati.timeout_ms / 1000;
    tv.tv_usec = D->ati.timeout_ms % 1000 * 1000;

    if (setsockopt(D->ati.socket, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv)))
        PRINTDEBUG(DBG_GENERIC_DRV, ("setsockopt for SO_RECVTIMEO failed with error: %m\n"));

    if (setsockopt(D->ati.socket, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv)))
        PRINTDEBUG(DBG_GENERIC_DRV, ("setsockopt for SO_SNDTIMEO failed with error: %m\n"));

    // printf("done constructed!\n");
}

void adwin_tcp_destroy()
{
    XTRACE("Calling destructor!\n");
    if (D->ati.req.data.as_void) free(D->ati.req.data.as_void);
    if (D->ati.resp.data.as_void) free(D->ati.resp.data.as_void);
    close(D->ati.socket);
    D->ati.connected = 0;
    D->ati.initialized =  0;
}

void init_ati(int32_t command)
{
    D->ati.command = command;

    if (D->ati.req.data.as_int32[SMALL_REQ_SIZE / 4] != 0x12348765) {
        XTRACE("Stack corruption req!\n");
    }
    if (D->ati.resp.data.as_int32[SMALL_REQ_SIZE / 4] != 0x22334455) {
        XTRACE("Stack corruption resp!\n");
    }
    D->ati.req.do_convert = ADWIN_HTONL;
    D->ati.req.used = 0;
    D->ati.resp.do_convert = ADWIN_NTOHL;
    D->ati.resp.used = 0;
    D->ati.status = -1111;
}

int32_t check_tcp_ret(int32_t ret, int32_t ref)
{
    if (ret == ref) return 0;

    set_last_error(k_error_tcp_timout_read);
    XTRACE("Socket-error: %m\n");
    close(D->ati.socket);
    D->ati.connected = 0;
    return 1;
}

int32_t tcp_read_response(adwin_tcp_info_t *ati, int32_t min_bytes, adwin_tcp_buffer_t *buf)
{
    adwin_tcp_header_t ath;
    adwin_tcp_buffer_t *resp = buf ? buf : &ati->resp;

try_again:
    ;

    int32_t ret = readn(ati->socket, (char *)&ath, sizeof(ath));
    if (check_tcp_ret(ret, sizeof(ath))) {
        XTRACE("readn header returned: %d\n", ret);
        set_last_error(k_error_tcp_error);
        return -1;
    }

    ath.magic_flag = ntohl(ath.magic_flag);
    if (check_tcp_ret(ath.magic_flag, ADWIN_TCP_MAGIC_FLAG_RESPONSE)) {
        XTRACE("magic flag error: 0x%08x instead of 0x%08x\n", ath.magic_flag, ADWIN_TCP_MAGIC_FLAG_RESPONSE);
        set_last_error(k_error_tcp_error);
        return -1;
    }
    ati->rx_pending = ntohl(ath.length);
    ati->status = ntohl(ath.command_or_status);
    if (ati->status == k_TryLater)
        goto try_again;

    set_last_error(ati->status);

    return tcp_read_pending(ati, min_bytes, resp);
}

void do_convert(adwin_tcp_buffer_t *buf)
{
    if (! buf) return;

    int32_t i, *p = buf->data.as_int32;

    switch (buf->do_convert) {
        case ADWIN_NTOHL:
            for (i = 0; i < buf->used / 4; i++, p++) {
                *p = ntohl(*p);
            }
            break;
        case ADWIN_HTONL:
            for (i = 0; i < buf->used / 4; i++, p++) {
                *p = htonl(*p);
            }
            break;
    }
}

int32_t tcp_write_request(adwin_tcp_info_t *ati, adwin_tcp_buffer_t *buf)
{
    adwin_tcp_buffer_t tmp = { 0, 0, 0, { NULL } };
    if (! buf) {
        buf = &tmp;
    }

    ati->req_count++;
    //XTRACE("REQ %d: cmd %d with %d bytes\n", ati->req_count, ati->command, req->used);
    int32_t req_len = ati->req.used + buf->used, bytes_current_transfer, bytes_already_send = 0, num_buffers = 2;

    adwin_tcp_header_t ath = {
        htonl(ADWIN_TCP_MAGIC_FLAG_REQUEST),  // todo: define as network byte order
        htonl(ati->req_count),
        htonl(req_len),
        htonl(ati->command)
    };

    do_convert(&ati->req);
    do_convert(buf);

    // sending more than 10MB in one transaction causes timeout issues!
    if (buf->used > ati->timeout_size) {
        bytes_current_transfer = ati->timeout_size;
    } else {
        bytes_current_transfer = buf->used;
    }

    if (buf->data.as_char) {
        num_buffers = 3;
    }
    struct iovec buffers[3] = { { (char *) &ath, sizeof(ath) },
        { ati->req.data.as_char, ati->req.used},
        { buf->data.as_char, bytes_current_transfer },
    };
    int ret = writev(ati->socket, buffers, num_buffers);
    bytes_already_send = bytes_current_transfer;

    if (check_tcp_ret(ret, sizeof(ath) + ati->req.used + bytes_current_transfer)) {
        PRINTDEBUG(DBG_GENERIC_DRV, ("WSASend returned %d\n", ret));
        set_last_error(k_error_tcp_error);
        return -1;
    }
    // transfer rest in MAX_DATA_PER_TRANSFER chunks
    while (bytes_already_send < buf->used) {
        if (bytes_already_send + ati->timeout_size <= buf->used) {
            bytes_current_transfer = ati->timeout_size;
        } else {
            bytes_current_transfer = buf->used - bytes_already_send;
        }
        ret = send(ati->socket, &buf->data.as_char[bytes_already_send], bytes_current_transfer, 0);
        if (check_tcp_ret(ret, bytes_current_transfer)) {
            set_last_error(k_error_tcp_error);
            return -1;
        }
        bytes_already_send += bytes_current_transfer;
    }

    return 0;
}

/* int32_t tcp_write_request(adwin_tcp_info_t *ati) { */
/*  adwin_tcp_buffer_t tmp = { 0, 0, 0, NULL }; */
/*  return tcp_write_request(ati, &tmp); */
/* } */

int32_t tcp_read_pending(adwin_tcp_info_t *ati, int32_t min_bytes, adwin_tcp_buffer_t *buf)
{
    int32_t bytes_current_transfer, bytes_already_read = 0;
    adwin_tcp_buffer_t *resp = buf ? buf : &ati->resp;

    if (resp->max < min_bytes) {
        XTRACE("error: not enough room in destination memory\n");
        set_last_error(k_error_tcp_error);
        return -1;
    }

    if (ati->rx_pending < min_bytes) {
        XTRACE("read pending wanted: %d, but only %d available\n", min_bytes, ati->rx_pending);
        set_last_error(k_error_tcp_timout_read);
        return -1;
    }

    while (min_bytes > bytes_already_read) {
        if (min_bytes > bytes_already_read + ati->timeout_size) {
            bytes_current_transfer = ati->timeout_size;
        } else {
            bytes_current_transfer = min_bytes - bytes_already_read;
        }
        int32_t ret = readn(ati->socket, &resp->data.as_char[bytes_already_read], bytes_current_transfer);
        if (check_tcp_ret(ret, bytes_current_transfer)) {
            XTRACE("readn body returned: %d expected %d\n", ret, bytes_current_transfer);
            return -1;
        }
        bytes_already_read += bytes_current_transfer;
    }

    resp->used = bytes_already_read;
    ati->rx_pending -= bytes_already_read;
    do_convert(resp);

    return 0;
}

/* int32_t tcp_read_pending(adwin_tcp_info_t *ati, int32_t min_bytes) { */
/*  return tcp_read_pending(ati, min_bytes, NULL); */
/* } */

int32_t tcp_discard_pending(adwin_tcp_info_t *ati)
{
    if (ati->rx_pending == 0) {
        return k_ok;
    }

    adwin_tcp_buffer_t tmp = { ati->rx_pending, 0, 0, { NULL } };
    tmp.data.as_void = malloc(ati->rx_pending);
    int32_t ret  = tcp_read_pending(ati, ati->rx_pending, &tmp);
    free(tmp.data.as_void);
    return ret;
}

void adwin_tcp_connect(void)
{
    socklen_t addrlen;
    int32_t ret;
    fd_set fds_read, fds_write, fds_except;
    struct timeval timeout;

    D->ati.connected = 0;

    XTRACE("%s: server_addr = %s, port = %d, password = '%s'\n", __FUNCTION__, inet_ntoa(D->ati.server_addr.sin_addr), D->port, D->passwd);

    addrlen = sizeof(struct sockaddr);

    ret = connect(D->ati.socket, (struct sockaddr *) & (D->ati.server_addr), addrlen);
    if (check_tcp_ret(0, ret)) {
        return;
    }

    // todo:
    /* if (ret == SOCKET_ERROR) { */
    /*      error_number = WSAGetLastError(); */
    /*      if (error_number == WSAEWOULDBLOCK) */
    /*          error_number = 0; */
    /*      else { */
    /*          XTRACE("connect-error (in line %d): %ld\n", __LINE__, WSAGetLastError()); */
    /*          return; */
    /*      } */
    /* } */

    FD_ZERO(&fds_read);
    FD_SET(D->ati.socket, &fds_read);
    FD_ZERO(&fds_write);
    FD_SET(D->ati.socket, &fds_write);
    FD_ZERO(&fds_except);
    FD_SET(D->ati.socket, &fds_except);

    /* Den gewählte timeout-Wert einsetzen */
    timeout.tv_sec = 1;     //<------------- Erstmal zum Testen ???????????????
    timeout.tv_usec = 0;

    /* Nun select aufrufen; dieses kehrt entweder nach Ablauf des Timeouts
     * zurück, oder wenn der Socket zum Schreiben bereit ist, was genau dann
     * passiert, wenn er erfolgreich verbunden wurde.
     */
    ret = select(D->ati.socket + 1, &fds_read, &fds_write, &fds_except, &timeout);
    if (ret == -1) {
        check_tcp_ret(0, 1);
        return;
    }

    /* Falls select zurückgekehrt ist, aber der zu verbindende Socket nicht
     * im Deskriptor-Set vorhanden ist, war das Verbinden in der gegebenen
     * Zeit nicht erfolgreich.
     */
    if (FD_ISSET(D->ati.socket, &fds_except)) {
        XTRACE("in fds_except\n");
        check_tcp_ret(0, 1);
        return;
    }
    if ((FD_ISSET(D->ati.socket, &fds_read) == 0) && (FD_ISSET(D->ati.socket, &fds_write) == 0)) {
        XTRACE("not in fds_read or fds_write\n");
        check_tcp_ret(0, 1);
        return;
    }

    /* Der Socket kann nun wieder blockierend gemacht werden */
    // opt = 0;
    // ioctlsocket(D->ati.socket, FIONBIO, &opt); todo blocking todo andere fiobio suchen!

    init_ati(k_Authenticate);
    D->ati.req.data.as_int32[0] = htonl(D->ati.dll_version);
    D->ati.req.data.as_int32[4] = htonl(D->ati.timeout_ms);
    D->ati.req.data.as_int32[5] = htonl(D->ati.timeout_size);
    strncpy(&D->ati.req.data.as_char[4], D->passwd, 10);
    D->ati.req.used = 24;
    D->ati.req.do_convert = 0;

    if (tcp_write_request(&D->ati, NULL)) {
        check_tcp_ret(0, 1);
        return;
    }

    if (tcp_read_response(&D->ati, 16, NULL)) {
        check_tcp_ret(0, 1);
        return;
    }

    D->ati.processortype = D->ati.resp.data.as_int32[0] + 9; // 0 = T9, ..., 3 = T12
    //connect_count = D->ati.resp.data.as_int32[1];
    D->ati.firmware_version = D->ati.resp.data.as_int32[2];
    D->ati.firmware_version_beta = D->ati.resp.data.as_int32[3];

    switch (D->ati.processortype) {
        case 9:
            XTRACE("processortype: T9 (%d)\n", D->ati.processortype);
            D->ati.processortype = PROCESSOR_T9;
            break;
        case 10:
            XTRACE("processortype: T10 (%d)\n", D->ati.processortype);
            D->ati.processortype = PROCESSOR_T10;
            break;
        case 11:
            XTRACE("processortype: T11 (%d)\n", D->ati.processortype);
            D->ati.processortype = PROCESSOR_T11;
            break;
        case 12:
            XTRACE("processortype: T12 (%d)\n", D->ati.processortype);
            D->ati.processortype = PROCESSOR_T12;
            break;
        default:
            XTRACE("unknown processortype: %d\n", D->ati.processortype);
            D->ati.processortype = -1;
    }
    XTRACE("connect_count: %d\n", connect_count);
    XTRACE("Version: %.2f, Beta: %.2f\n", D->ati.firmware_version / 100., D->ati.firmware_version_beta / 100.);

    if (D->ati.error_number == k_ok) {
        D->ati.connected = 1;
    } else {
        // XTRACE("Status: %d\n", D->ati.status);
        close(D->ati.socket);
    }
}

int32_t bytes_per_value(adwin_data_type_t adt)
{
    int32_t ADWIN_DATA_BYTES_PER_VALUE[] = {0, 1, 2, 4, 4, 4, 8, 8, 1};
    if ((adt < 0) || (adt > sizeof(ADWIN_DATA_BYTES_PER_VALUE) / sizeof(ADWIN_DATA_BYTES_PER_VALUE[0]))) {
        XTRACE("Invalid data type %d\n", adt);
        return -1;
    }
    return ADWIN_DATA_BYTES_PER_VALUE[adt];
}

int32_t map_error_code(int32_t adwin_error)
{
    switch (adwin_error) {
        case 5:
        case -1:
            return k_error_tcp_timout_read;
        default:
            return adwin_error;
    }
}


/*int32_t map_error_code(int32_t adwin_error)
{
    if (adwin_error)
        printf("################# map_error_code %d\n", adwin_error);
    switch (adwin_error) {
        case 0:
            return k_ok;
        case 1:
            return k_timeout_to_link;
        case 2:
            return k_timeout_from_link;
        case 5:
        case -1:
            return k_error_tcp_timout_read;
        case -5:
            return k_WrongPassword;
        //case -10: try_later
        case -15:
            return k_DateiFehler;
        case -31:
            return k_DataZuKlein;
        case -32:
            return k_FifoLessData;
        case -35:
            return k_WrongBinfile;
        case -100:
            return k_MemError;
        default:
            return adwin_error;
    }
}*/

void set_last_error(int32_t new_error_number)
{
    if (new_error_number || D->ati.error_number) {
        XTRACE("Setting error_number: was %d, want %d (0x%08x), after mapping: %d (0x%08x)\n",
               D->ati.error_number, new_error_number, new_error_number,  map_error_code(new_error_number), map_error_code(new_error_number));
    }
    D->ati.error_number = map_error_code(new_error_number);
}

//typedef union cmd4p4_s {
//  int32_t i32[4];
//  float f[4];
//  double d[2];
//} cmd4p4_t;
void command_4p4(int32_t to_values[4], int32_t from_values[4])
{
    adwin_tcp_buffer_t req = { 16, 16, ADWIN_HTONL, { to_values } }, resp = { 16, 0, ADWIN_NTOHL, { from_values } } ;

    init_ati(k_3plus1);

    if (tcp_write_request(&D->ati, &req)) return;

    if (tcp_read_response(&D->ati, 16, &resp)) return;
}

void boot_or_loadprocess(const char *filename, int32_t command)
{
    FILE *fd = NULL;
    struct stat filestat;
    int32_t ret;
    adwin_tcp_buffer_t req = { 0, 0, 0, { NULL } };

    init_ati(command);

    if ((fd = fopen(filename, "rb")) == NULL) {
        XTRACE("%s\n", filename);
        perror("fopen");
        set_last_error(k_File_not_found);
        goto free_return;
    }

    if (stat(filename, &filestat)) {
        perror("stat");
        set_last_error(k_File_not_found);
        goto free_return;
    }

    req.max = req.used = filestat.st_size;
    req.data.as_void = malloc(filestat.st_size);

    ret = fread(req.data.as_void, sizeof(char), filestat.st_size, fd);
    if (ret != filestat.st_size) {
        perror("fread");
        set_last_error(k_WrongBinFile);
        goto free_return;
    }

    if (tcp_write_request(&D->ati, &req)) goto free_return;

    D->ati.resp.do_convert = 0;
    if (tcp_read_response(&D->ati, 0, NULL)) goto free_return;

free_return:
    if (req.data.as_void)
        free(req.data.as_void);
    if (fd)
        fclose(fd);
}

void adwin_tcp_boot(const char *filename)
{
    boot_or_loadprocess(filename, k_Boot);
}

void adwin_tcp_load_process(const char *filename)
{
    boot_or_loadprocess(filename, k_LadeBin);
}

void adwin_tcp_start_process(int32_t processnumber)
{
    int32_t from_values[4], to_values[4] = { k_Start, processnumber, 0, 0 };

    command_4p4(to_values, from_values);
}

void adwin_tcp_stop_process(int32_t processnumber)
{
    int32_t from_values[4], to_values[4] = { k_Stop, processnumber, 0, 0 };

    command_4p4(to_values, from_values);
}

void adwin_tcp_clear_process(int32_t processnumber)
{

    int32_t from_values[4], to_values[4] = { k_ClearProcess, processnumber, 0, 0 };

    command_4p4(to_values, from_values);
}
// TODO: stimmen die alle auch fuer prozesse > 10???
int32_t adwin_tcp_process_status(int32_t processnumber)
{
    return adwin_tcp_get_par(-100 + processnumber);
}

int32_t adwin_tcp_get_processdelay(int32_t processnumber)
{
    return adwin_tcp_get_par(-90 + processnumber);
}

void adwin_tcp_set_processdelay(int32_t processnumber, int32_t delay)
{
    adwin_tcp_set_par(-90 + processnumber, delay);
}

int32_t adwin_tcp_testversion(int32_t values[4])
{
    init_ati(k_TestVersion);
    adwin_tcp_buffer_t resp = { 16, 0, ADWIN_HTONL, { values } };

    if (tcp_write_request(&D->ati, NULL)) return -1;

    if (tcp_read_response(&D->ati, 16, &resp)) return -1;

    return 0;
}

float adwin_tcp_workload(void)
{
    float workload = -1;

    init_ati(k_Auslastung);

    if (tcp_write_request(&D->ati, NULL)) return -1;

    if (tcp_read_response(&D->ati, 16, NULL)) return -1;

    if (D->ati.status) return -1;

    if (D->ati.processortype != PROCESSOR_T12) {
        uint32_t workload_calculated_by_dsp, val1, t1;
        float help1;

        val1 = D->ati.resp.data.as_uint32[0];
        t1 = D->ati.resp.data.as_uint32[1];
        workload_calculated_by_dsp = D->ati.resp.data.as_int32[2]; // immer 0?

        //XTRACE("values: %f, %f, %d\n", val1, t1, val3);
        if (workload_calculated_by_dsp == 0) {
            if (t1 == 0) {
                return 100;
            } else {
                help1 = (float)val1 / (float)t1;
                //XTRACE("xxx %f %f %f\n", (float)val1, (float)t1, help1);
                if (help1 == 0) {
                    return 100.0;
                }
                if (D->ati.processortype == PROCESSOR_T11)
                    workload = (100.f * (((1.f / 9.f) - help1) * 9.f));
                else // T9
                    workload = (100.f * ((0.25f - help1) / 0.25f));
            }
        } else
            workload = val1;
    } else
        workload = D->ati.resp.data.as_int32[0];

    if (workload > 100)
        return 100;
    if (workload < 0)
        return 0;
    return workload;
}

int32_t adwin_tcp_free_mem(int32_t mem_spec)
{
    int32_t from_values[4], to_values[4] = { k_MemoryAll, mem_spec, 0, 0};

    command_4p4(to_values, from_values);

    return from_values[0];
}

void adwin_tcp_peek(int32_t address, int32_t count, int32_t *pc_array)
{
    init_ati(k_GetMemoryDSP);
    D->ati.req.used = 8;
    D->ati.req.data.as_int32[0] = address;
    D->ati.req.data.as_int32[1] = count;

    if (tcp_write_request(&D->ati, NULL)) return;

    adwin_tcp_buffer_t resp;
    resp.data.as_int32 = pc_array;
    resp.max = count * 4;
    resp.used = 0;
    resp.do_convert = 0;

    // get data
    if (tcp_read_response(&D->ati, resp.max, &resp)) return;

    // get status
    if (tcp_read_pending(&D->ati, 4, NULL)) return;

    set_last_error(D->ati.resp.data.as_int32[0]);
}

int32_t adwin_tcp_get_par(int32_t par_no)
{
    int32_t from_values[4], to_values[4] = { k_GetPar, par_no + 1000, 0, 0};

    command_4p4(to_values, from_values);

    return from_values[0];
}

void adwin_tcp_set_par(int32_t par_no, int32_t value)
{
    int32_t from_values[4], to_values[4] = { k_SetPar, par_no + 1000, value, 0};

    command_4p4(to_values, from_values);
}

float adwin_tcp_get_fpar(int32_t par_no)
{
    int32_t from_values[4], to_values[4] = { k_GetPar, par_no + 1100, 0, 0};

    command_4p4(to_values, from_values);

    // maybe todo:
    // cmd4p4_t resp;
    // return resp.f[0];

    void *tmp = &from_values[0];
    return *(float *)tmp;
}

int32_t adwin_tcp_set_value(int32_t address, void *value, int type_size)
{
    int32_t from_values[4];

	if (type_size == 8)
	{
		int32_t to_values[4] = {k_SetMem_8, address, 0, 0};
		memcpy(&to_values[2], value, 8);
		command_4p4(to_values, from_values);
	}
	else
	{
		int32_t to_values[4] = {k_SetMem_To_4, address, *(long *) value, type_size};
		memcpy(&to_values[2], value, type_size);
		command_4p4(to_values, from_values);
	}

	return from_values[0];
}

void adwin_tcp_set_fpar(int32_t par_no, float value)
{
    void *tmp = &value;
    int32_t from_values[4], to_values[4] = { k_SetPar, par_no + 1100, *(int32_t *)tmp, 0 };

    // maybe todo
    // cmd4p4_t req = {{ k_SetPar, par_no + 1100, 0, 0 }};
    // req.f[3] = value;

    command_4p4(to_values, from_values);
}

double adwin_tcp_get_fpar_double(int32_t par_no)
{
    if (D->ati.processortype != PROCESSOR_T12)
        return adwin_tcp_get_fpar(par_no);

    int32_t from_values[4], to_values[4] = { k_GetPar, par_no + 1200, 0, 0 };
    command_4p4(to_values, from_values);

    void *tmp = &from_values[0];
    return *(double *)tmp;
}

void adwin_tcp_set_fpar_double(int32_t par_no, double value)
{
    int32_t to_values[4];
    int32_t from_values[4];
    int32_t dummy;

    if (D->ati.processortype != PROCESSOR_T12) {
        adwin_tcp_set_fpar(par_no, (float)value);
        return;
    }

    to_values[0] = k_SetPar;
    to_values[1] = par_no + 1200;
    memcpy(&to_values[2], &value, sizeof(double));

    dummy = to_values[2];
    to_values[2] = to_values[3];
    to_values[3] = dummy;

    command_4p4(to_values, from_values);
}

void adwin_tcp_get_par_block(int32_t start, int32_t count, int32_t *pars)
{
    init_ati(k_GetParAll);
    D->ati.req.data.as_int32[0] = start;
    D->ati.req.data.as_int32[1] = count;
    D->ati.req.used = 8;

    if (tcp_write_request(&D->ati, NULL)) return;

    adwin_tcp_buffer_t resp;
    resp.data.as_int32 = pars;
    resp.max = count * 4;
    resp.used = 0;
    resp.do_convert = 0;
    if (tcp_read_response(&D->ati, resp.max, &resp)) return;
}

void adwin_tcp_get_par_all(int32_t *pars)
{
    adwin_tcp_get_par_block(1, 80, pars);
}

void adwin_tcp_get_fpar_block(int32_t start, int32_t count, float *fpars)
{
    adwin_tcp_get_par_block(100 + start, count, (int32_t *) fpars);
}

void adwin_tcp_get_fpar_all(float *fpars)
{
    adwin_tcp_get_par_block(101, 80, (int32_t *) fpars);
}

void adwin_tcp_get_fpar_block_double(int32_t start, int32_t count, double *dpars)
{
    float *fpars;
    int i;

    if (D->ati.processortype != PROCESSOR_T12) {
        fpars = (float *) malloc(count * 4);
        adwin_tcp_get_par_block(100 + start, count, (int32_t *) fpars);
        for (i = 0; i < count; i++)
            dpars[i] = fpars[i];
        free(fpars);
    } else
        adwin_tcp_get_par_block(200 + start, count * 2, (int32_t *) dpars);
}

void adwin_tcp_get_fpar_all_double(double *dpars)
{
    float *fpars;
    int i;

    if (D->ati.processortype != PROCESSOR_T12) {
        fpars = (float *) malloc(80 * 4);
        adwin_tcp_get_par_block(101, 80, (int32_t *) fpars);
        for (i = 0; i < 80; i++)
            dpars[i] = fpars[i];
        free(fpars);
    } else
        adwin_tcp_get_par_block(201, 160, (int32_t *) dpars);
}

adwin_data_type_t adwin_tcp_get_data_type(int32_t datanumber)
{
    init_ati(k_GetDataType);
    D->ati.req.data.as_int32[0] = datanumber;
    D->ati.req.used = 4;

    if (tcp_write_request(&D->ati, NULL)) return ADWIN_DATA_TYPE_INVALID;

    if (tcp_read_response(&D->ati, 4, NULL)) return ADWIN_DATA_TYPE_INVALID;

    return (adwin_data_type_t)D->ati.resp.data.as_int32[0];
}

int32_t adwin_tcp_get_data_length(int32_t datanumber)
{
    int32_t from_values[4], to_values[4] = { k_GetDataLength, datanumber, 0, 0};

    command_4p4(to_values, from_values);

    return from_values[0];
}

void adwin_tcp_clear_data(int32_t datanumber)
{
    int32_t from_values[4], to_values[4] = { k_Clear_Data, datanumber, 0, 0 };

    command_4p4(to_values, from_values);
}


void convert_data(int from_type, int to_type, const void *from_data, void *to_data, int32_t count)
{
    int32_t i;
#if 0
    ADWIN_DATA_TYPE_INVALID = -1,
    ADWIN_BYTE = 1,    // int_8
    ADWIN_SHORT = 2,   // int_16
    ADWIN_LONG = 3,    // int_32
    ADWIN_INTEGER = 4, // int32, usually no used
    ADWIN_FLOAT = 5,   // float32
    ADWIN_DOUBLE = 6,  // float64
    ADWIN_INT64 = 7,   // int_64
    ADWIN_STRING = 8,  // int_8, only T12
#endif

    //printf("%s: from_type: %d, to_type: %d, count: %d\n", __FUNCTION__, from_type, to_type, count);
    //XTRACE("convert from %d to %d\n", from_type, to_type);

    switch (to_type) {
        case ADWIN_DATA_TYPE_INVALID:
            // ?
            XTRACE("invalid data type in convert_data!\n");
            break;
        case ADWIN_BYTE:
        case ADWIN_STRING: {
            int8_t *p_int8 = (int8_t *)to_data;
            switch (from_type) {
                case ADWIN_BYTE:
                case ADWIN_STRING:
                    XTRACE("Useless data conversion!\n");
                    for (i = 0; i < count; i++)
                        p_int8[i] = ((int8_t *) from_data)[i];
                    break;
                case ADWIN_SHORT:
                    for (i = 0; i < count; i++)
                        p_int8[i] = ((int16_t *) from_data)[i];
                    break;
                case ADWIN_LONG:
                case ADWIN_INTEGER:
                    for (i = 0; i < count; i++)
                        p_int8[i] = ((int32_t *) from_data)[i];
                    break;
                case ADWIN_FLOAT:
                    for (i = 0; i < count; i++)
                        p_int8[i] = ((float *) from_data)[i];
                    break;
                case ADWIN_DOUBLE:
                    for (i = 0; i < count; i++)
                        p_int8[i] = ((double *) from_data)[i];
                    break;
                case ADWIN_INT64:
                    for (i = 0; i < count; i++)
                        p_int8[i] = ((int64_t *) from_data)[i];
                    break;
            }
            if (to_type == ADWIN_STRING) {
                p_int8[count - 1] = 0; // ensure 0 termination
            }
            break;
        }
        case ADWIN_SHORT: {
            int16_t *p_int16 = (int16_t *)to_data;
            switch (from_type) {
                case ADWIN_BYTE:
                case ADWIN_STRING:
                    for (i = 0; i < count; i++)
                        p_int16[i] = ((int8_t *) from_data)[i];
                    break;
                case ADWIN_SHORT:
                    XTRACE("Useless data conversion!\n");
                    for (i = 0; i < count; i++)
                        p_int16[i] = ((int16_t *) from_data)[i];
                    break;
                case ADWIN_LONG:
                case ADWIN_INTEGER:
                    for (i = 0; i < count; i++)
                        p_int16[i] = ((int32_t *) from_data)[i];
                    break;
                case ADWIN_FLOAT:
                    for (i = 0; i < count; i++)
                        p_int16[i] = ((float *) from_data)[i];
                    break;
                case ADWIN_DOUBLE:
                    for (i = 0; i < count; i++)
                        p_int16[i] = ((double *) from_data)[i];
                    break;
                case ADWIN_INT64:
                    for (i = 0; i < count; i++)
                        p_int16[i] = ((int64_t *) from_data)[i];
                    break;
            }
            break;
        }
        case ADWIN_LONG:
        case ADWIN_INTEGER: {
            int32_t *p_int32 = (int32_t *)to_data;
            switch (from_type) {
                case ADWIN_BYTE:
                case ADWIN_STRING:
                    for (i = 0; i < count; i++)
                        p_int32[i] = ((int8_t *) from_data)[i];
                    break;
                case ADWIN_SHORT:
                    for (i = 0; i < count; i++)
                        p_int32[i] = ((int16_t *) from_data)[i];
                    break;
                case ADWIN_LONG:
                case ADWIN_INTEGER:
                    XTRACE("Useless data conversion!\n");
                    for (i = 0; i < count; i++)
                        p_int32[i] = ((int32_t *) from_data)[i];
                    break;
                case ADWIN_FLOAT:
                    for (i = 0; i < count; i++)
                        p_int32[i] = ((float *) from_data)[i];
                    break;
                case ADWIN_DOUBLE:
                    for (i = 0; i < count; i++)
                        p_int32[i] = ((double *) from_data)[i];
                    break;
                case ADWIN_INT64:
                    for (i = 0; i < count; i++)
                        p_int32[i] = ((int64_t *) from_data)[i];
                    break;
            }
            break;
        }
        case ADWIN_FLOAT: {
            float *p_float = (float *)to_data;
            switch (from_type) {
                case ADWIN_BYTE:
                case ADWIN_STRING:
                    for (i = 0; i < count; i++)
                        p_float[i] = ((int8_t *) from_data)[i];
                    break;
                case ADWIN_SHORT:
                    for (i = 0; i < count; i++)
                        p_float[i] = ((int16_t *) from_data)[i];
                    break;
                case ADWIN_LONG:
                case ADWIN_INTEGER:
                    for (i = 0; i < count; i++)
                        p_float[i] = ((int32_t *) from_data)[i];
                    break;
                case ADWIN_FLOAT:
                    XTRACE("Useless data conversion!\n");
                    for (i = 0; i < count; i++)
                        p_float[i] = ((float *) from_data)[i];
                    break;
                case ADWIN_DOUBLE:
                    for (i = 0; i < count; i++)
                        p_float[i] = ((double *) from_data)[i];
                    break;
                case ADWIN_INT64:
                    for (i = 0; i < count; i++)
                        p_float[i] = ((int64_t *) from_data)[i];
                    break;
            }
            break;
        }
        case ADWIN_DOUBLE: {
            double *p_double = (double *)to_data;
            switch (from_type) {
                case ADWIN_BYTE:
                case ADWIN_STRING:
                    for (i = 0; i < count; i++)
                        p_double[i] = ((int8_t *) from_data)[i];
                    break;
                case ADWIN_SHORT:
                    for (i = 0; i < count; i++)
                        p_double[i] = ((int16_t *) from_data)[i];
                    break;
                case ADWIN_LONG:
                case ADWIN_INTEGER:
                    for (i = 0; i < count; i++)
                        p_double[i] = ((int32_t *) from_data)[i];
                    break;
                case ADWIN_FLOAT:
                    for (i = 0; i < count; i++)
                        p_double[i] = ((float *) from_data)[i];
                    break;
                case ADWIN_DOUBLE:
                    XTRACE("Useless data conversion!\n");
                    for (i = 0; i < count; i++)
                        p_double[i] = ((double *) from_data)[i];
                    break;
                case ADWIN_INT64:
                    for (i = 0; i < count; i++)
                        p_double[i] = ((int64_t *) from_data)[i];
                    break;
            }
            break;
        }
        case ADWIN_INT64: {
            int64_t *p_int64 = (int64_t *)to_data;
            switch (from_type) {
                case ADWIN_BYTE:
                case ADWIN_STRING:
                    for (i = 0; i < count; i++)
                        p_int64[i] = ((int8_t *) from_data)[i];
                    break;
                case ADWIN_SHORT:
                    for (i = 0; i < count; i++)
                        p_int64[i] = ((int16_t *) from_data)[i];
                    break;
                case ADWIN_LONG:
                case ADWIN_INTEGER:
                    for (i = 0; i < count; i++)
                        p_int64[i] = ((int32_t *) from_data)[i];
                    break;
                case ADWIN_FLOAT:
                    for (i = 0; i < count; i++)
                        p_int64[i] = ((float *) from_data)[i];
                    break;
                case ADWIN_DOUBLE:
                    for (i = 0; i < count; i++)
                        p_int64[i] = ((double *) from_data)[i];
                    break;
                case ADWIN_INT64:
                    for (i = 0; i < count; i++)
                        p_int64[i] = ((int64_t *) from_data)[i];
                    break;
            }
            break;
        }
        break;
    }
}

void adwin_tcp_get_data_or_fifo(adwin_system_data_type_t asdt, int32_t datanumber, void *data, int32_t start, int32_t count, int32_t get_fifo)
{
    adwin_system_data_type_t _asdt;
    int32_t count_b;
    void *buf = NULL;
    adwin_tcp_buffer_t resp;

    if (count < 1) {
        // TODO: Fehlercode
        set_last_error(k_WrongSize);
        goto free_return;
    }

    init_ati(get_fifo ? k_GetFifo : k_GetData);
    D->ati.req.used = 12;
    D->ati.req.data.as_int32[0] = datanumber;
    D->ati.req.data.as_int32[1] = start;
    D->ati.req.data.as_int32[2] = count;

    if (tcp_write_request(&D->ati, NULL)) goto free_return;

    // get initial response
    if (tcp_read_response(&D->ati, 12, NULL)) goto free_return;

    if (D->ati.status) goto free_return;

    _asdt = (adwin_data_type_t)D->ati.resp.data.as_int32[0];
    count_b = count * bytes_per_value(_asdt);

    if (bytes_per_value(_asdt) < 0) {
        D->ati.status = k_Mem_Error;
        tcp_discard_pending(&D->ati);
        goto free_return;
    }

    if (asdt != _asdt) {
        resp.data.as_void = buf = malloc(count_b);
        if (resp.data.as_void == NULL) {
            set_last_error(k_Mem_Error);
            perror("malloc");
            goto free_return;
        }
    } else {
        resp.data.as_void = data;
    }

    // get data
    resp.max = count_b;
    resp.do_convert = 0;
    resp.used = 0;
    if (tcp_read_pending(&D->ati, count_b, &resp)) goto free_return;
    // get status
    if (tcp_read_pending(&D->ati, 4, NULL)) goto free_return;

    if (D->ati.resp.data.as_int32[0] != k_ok) goto free_return;

    if (_asdt != asdt) {
        convert_data(_asdt, asdt, resp.data.as_void, data, count);
    }

free_return:
    D->fehlercode = map_status_to_errorcode(D->ati.status);
    set_last_error(D->fehlercode);
    if (buf) free(buf);
}

void adwin_tcp_get_data(adwin_system_data_type_t asdt, int32_t datanumber, void *data, int32_t start, int32_t count)
{
    adwin_tcp_get_data_or_fifo(asdt, datanumber, data, start, count, 0);
}
void adwin_tcp_get_data_long(int32_t datanumber, void *data, int32_t start, int32_t count)
{
    adwin_tcp_get_data(ADWIN_LONG, datanumber, data, start, count);
}

void adwin_tcp_get_data_float(int32_t datanumber, void *data, int32_t start, int32_t count)
{
    adwin_tcp_get_data(ADWIN_FLOAT, datanumber, data, start, count);
}

void adwin_tcp_get_data_double(int32_t datanumber, void *data, int32_t start, int32_t count)
{
    adwin_tcp_get_data(ADWIN_DOUBLE, datanumber, data, start, count);
}

void adwin_tcp_get_data_string(int32_t datanumber, char *c_str, int32_t max_count)
{
    int32_t val, i, strlen, *str;

    if (D->ati.processortype == PROCESSOR_T12) {
        adwin_tcp_get_data(ADWIN_STRING, datanumber, c_str, 0, max_count);
        c_str[max_count - 1] = 0;
    } else {
        str = (int32_t *)malloc(4 * (max_count + 1));
        adwin_tcp_get_data(ADWIN_LONG, datanumber, str, 1, max_count + 1);
        strlen = str[0];
        if (strlen >= max_count)
            strlen = max_count - 1;
        for (i = 0; i < strlen; i++) {
            val = str[i + 1];
            if ((val > 0) && (val < 256))
                c_str[i] = val;
            else {
                c_str[i] = 0;
                break;
            }
            c_str[i + 1] = 0; // todo
        }
        free(str);
    }
}

int32_t adwin_tcp_string_length(int32_t datanumber)
{
    int32_t from_values[4];

    if (D->ati.processortype == PROCESSOR_T12) {
        int32_t to_values[4] = { k_GetDataStringSize, datanumber, 0, 0 };

        command_4p4(to_values, from_values);

        return from_values[0];
    } else {
        adwin_tcp_get_data(ADWIN_LONG, datanumber, from_values, 1, 1);

        return from_values[0];
    }
}

void adwin_tcp_set_data_or_fifo(adwin_data_type_t pc_datatype, int32_t datanumber, const void *data, int32_t start, int32_t count, int32_t set_fifo)
{
    int32_t count_b;
    adwin_data_type_t adt;
    void *buf = NULL;
    int32_t data_size, is_fifo, fifo_full;
    adwin_tcp_buffer_t req;

    if (count < 1) {
        // TODO: Fehlercode
        set_last_error(k_WrongSize);
        goto free_return;
    }

    adwin_tcp_get_data_info(datanumber, (int32_t *)&adt, &data_size, &is_fifo, &fifo_full); // todo: cast entfernen!
    if (D->ati.error_number) goto free_return;

    if (data_size < start - 1 + count) {
        set_last_error(k_Data_is_to_small);
        goto free_return;
    }
    // for T12, we can already do some more checks
    if (D->ati.processortype == PROCESSOR_T12) {
        if (is_fifo && set_fifo && (count > data_size - fifo_full)) {
            set_last_error(k_Fifo_is_to_small);
            goto free_return;
        }
        if (set_fifo && ! is_fifo) {
            set_last_error(k_Fifo_is_to_small);
            goto free_return;
        }
    }

    // bytecount depends on datatype
    count_b = count * bytes_per_value(adt);
    init_ati(set_fifo ? k_SetFifo : k_SetData);
    D->ati.req.data.as_int32[0] = datanumber;
    D->ati.req.data.as_int32[1] = start; // not used for fifo
    D->ati.req.data.as_int32[2] = count;
    D->ati.req.data.as_int32[3] = adt;
    D->ati.req.used = 16;
    req.do_convert = 0;
    req.used = count_b;

    if (adt != pc_datatype) {
        req.data.as_void = buf = malloc(count_b);
        if (req.data.as_void == NULL) {
            set_last_error(k_Mem_Error);
            perror("malloc");
            goto free_return;
        }

        convert_data(pc_datatype, adt, data, req.data.as_void, count);
    } else {
        req.data.as_const_void = data;
    }

    if (tcp_write_request(&D->ati, &req)) goto free_return;

    if (tcp_read_response(&D->ati, 0, NULL)) goto free_return;

    if (D->ati.status != k_ok) goto free_return;
free_return:
    if (buf) free(buf);
}

void adwin_tcp_set_data(adwin_data_type_t pc_datatype, int32_t datanumber, const void *data, int32_t start, int32_t count)
{
    adwin_tcp_set_data_or_fifo(pc_datatype, datanumber, data, start, count, 0);
}

void adwin_tcp_set_data_long(int32_t datanumber, const void *data, int32_t start, int32_t count)
{
    adwin_tcp_set_data(ADWIN_LONG, datanumber, data, start, count);
}

void adwin_tcp_set_data_float(int32_t datanumber, const void *data, int32_t start, int32_t count)
{
    adwin_tcp_set_data(ADWIN_FLOAT, datanumber, data, start, count);
}

void adwin_tcp_set_data_double(int32_t datanumber, const void *data, int32_t start, int32_t count)
{
    adwin_tcp_set_data(ADWIN_DOUBLE, datanumber, data, start, count);
}

void adwin_tcp_set_data_string(int32_t datanumber, const char *c_str, int32_t max_count)
{
    int32_t length = strlen(c_str);
    if (length > max_count) {
        length = max_count;
    }
    if (D->ati.processortype == PROCESSOR_T12) {
        adwin_tcp_set_data(ADWIN_STRING, datanumber, c_str, 0, length + 1);
    } else {
        int32_t *data = (int32_t *)malloc((length + 2) * sizeof(int32_t));
        int32_t i;

        data[0] = length;
        for (i = 0; i < length; i++) {
            data[i + 1] = c_str[i];
        }
        data[length + 1] = 0;
        adwin_tcp_set_data(ADWIN_LONG, datanumber, data, 1, length + 2);
        free(data);
    }
}

void adwin_tcp_get_fifo(adwin_data_type_t pc_datatype, int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_get_data_or_fifo(pc_datatype, datanumber, data, 0,  count, 1);
}

void adwin_tcp_get_fifo_long(int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_get_fifo(ADWIN_LONG, datanumber, data, count);
}

void adwin_tcp_get_fifo_float(int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_get_fifo(ADWIN_FLOAT, datanumber, data, count);
}

void adwin_tcp_get_fifo_double(int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_get_fifo(ADWIN_DOUBLE, datanumber, data, count);
}

void adwin_tcp_set_fifo(adwin_data_type_t pc_datatype, int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_set_data_or_fifo(pc_datatype, datanumber, data, 0, count, 1);
}

void adwin_tcp_set_fifo_long(int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_set_fifo(ADWIN_LONG, datanumber, data, count);
}

void adwin_tcp_set_fifo_float(int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_set_fifo(ADWIN_FLOAT, datanumber, data, count);
}

void adwin_tcp_set_fifo_double(int32_t datanumber, void *data, int32_t count)
{
    adwin_tcp_set_fifo(ADWIN_DOUBLE, datanumber, data, count);
}

int32_t adwin_tcp_fifo_full(int32_t datanumber)
{
    int32_t from_values[4], to_values[4] = { k_GetFifoCount, datanumber, 0, 0};

    command_4p4(to_values, from_values);

    return from_values[0];
}

int32_t adwin_tcp_fifo_empty(int32_t datanumber)
{
    int32_t from_values[4], to_values[4] = { k_GetFifoEmpty, datanumber, 0, 0 };

    command_4p4(to_values, from_values);

    return from_values[0];
}


void adwin_tcp_fifo_clear(int32_t datanumber)
{
    int32_t from_values[4], to_values[4] = { k_ClearFifo, datanumber, 0, 0 };

    command_4p4(to_values, from_values);
}

void adwin_tcp_create_data(int32_t datanumber, int32_t as_fifo, int32_t datasize, int32_t datatype, int32_t bank)
{
    init_ati(k_CreateData);
    D->ati.req.data.as_int32[0] = datanumber;
    D->ati.req.data.as_int32[1] = as_fifo;
    D->ati.req.data.as_int32[2] = datasize;
    D->ati.req.data.as_int32[3] = datatype;
    D->ati.req.data.as_int32[4] = bank;
    D->ati.req.used = 20;

    if (tcp_write_request(&D->ati, NULL)) return;

    if (tcp_read_response(&D->ati, 0, NULL)) return;
}

void adwin_tcp_get_data_info(int32_t datanumber, int32_t *datatype, int32_t *datasize, int32_t *isfifo, int32_t *fifo_full)
{
    init_ati(k_GetDataInfo);
    D->ati.req.data.as_int32[0] = datanumber;
    D->ati.req.used = 4;

    if (tcp_write_request(&D->ati, NULL)) return;

    if (tcp_read_response(&D->ati, 16, NULL)) return;

    // datatype and datasize are 0 data has not been allocated yet
    *datatype = D->ati.resp.data.as_int32[0];
    *datasize = D->ati.resp.data.as_int32[1];
    *isfifo = D->ati.resp.data.as_int32[2];
    *fifo_full = D->ati.resp.data.as_int32[3];
}

processor_type_t adwin_tcp_get_processor_type()
{
    return D->ati.processortype;
}

ssize_t readn(int fd, void *vptr, size_t n)
{
    size_t  nleft;
    ssize_t nread;
    char   *ptr;

    ptr = vptr;
    nleft = n;
    while (nleft > 0) {
        if ( (nread = read(fd, ptr, nleft)) < 0) {
            if (errno == EINTR)
                nread = 0;      /* and call read() again */
            else {
                PRINTDEBUG(DBG_GENERIC_DRV, ("readn returned -1 when reading %zu bytes\n", n));
                return (-1);
            }
        } else if (nread == 0)
            break;              /* EOF */
        nleft -= nread;
        ptr += nread;
    }
    return (n - nleft);         /* return >= 0 */
}

ssize_t writen(int fd, const void *vptr, size_t n)
{
    size_t nleft;
    ssize_t nwritten;
    const char *ptr;

    ptr = vptr;
    nleft = n;
    while (nleft > 0) {
        if ( (nwritten = write(fd, ptr, nleft)) <= 0) {
            if (nwritten < 0 && errno == EINTR)
                nwritten = 0;   /* and call write() again */
            else {
                PRINTDEBUG(DBG_GENERIC_DRV, ("writen returned -1 when reading %zu bytes\n", n));
                return (-1);    /* error */
            }
        }
        nleft -= nwritten;
        ptr += nwritten;
    }
    return (n);
}
